/*
 *  TOPPERS/ASP Kernel
 *      Toyohashi Open Platform for Embedded Real-Time Systems/
 *      Advanced Standard Profile Kernel
 * 
 *  Copyright (C) 2008-2011 by Embedded and Real-Time Systems Laboratory
 *              Graduate School of Information Science, Nagoya Univ., JAPAN
 * 
 *  L쌠҂́Cȉ(1)`(4)̏𖞂ꍇɌC{\tgEF
 *  Ai{\tgEFAς̂܂ށDȉjgpEE
 *  ρEĔzziȉCpƌĂԁj邱Ƃ𖳏ŋD
 *  (1) {\tgEFA\[XR[ȟ`ŗpꍇɂ́CL̒
 *      \C̗pщL̖ۏ؋K肪Ĉ܂܂̌`Ń\[
 *      XR[hɊ܂܂Ă邱ƁD
 *  (2) {\tgEFACCu`ȂǁC̃\tgEFAJɎg
 *      pł`ōĔzzꍇɂ́CĔzzɔhLgip
 *      ҃}jAȂǁjɁCL̒쌠\C̗pщL
 *      ̖ۏ؋Kfڂ邱ƁD
 *  (3) {\tgEFAC@ɑgݍނȂǁC̃\tgEFAJɎg
 *      płȂ`ōĔzzꍇɂ́Ĉꂩ̏𖞂
 *      ƁD
 *    (a) ĔzzɔhLgip҃}jAȂǁjɁCL̒
 *        쌠\C̗pщL̖ۏ؋Kfڂ邱ƁD
 *    (b) Ĕzž`ԂCʂɒ߂@ɂāCTOPPERSvWFNg
 *        񍐂邱ƁD
 *  (4) {\tgEFA̗pɂ蒼ړI܂͊ԐړIɐ邢Ȃ鑹
 *      QCL쌠҂TOPPERSvWFNgƐӂ邱ƁD
 *      ܂C{\tgEFÃ[U܂̓Gh[ÛȂ闝
 *      RɊÂCL쌠҂TOPPERSvWFNg
 *      Ɛӂ邱ƁD
 * 
 *  {\tgEFÁCۏ؂Œ񋟂Ă̂łDL쌠҂
 *  TOPPERSvWFNǵC{\tgEFAɊւāC̎gpړI
 *  ɑ΂K܂߂āCȂۏ؂sȂD܂C{\tgEF
 *  A̗pɂ蒼ړI܂͊ԐړIɐȂ鑹QɊւĂC
 *  ̐ӔC𕉂ȂD
 * 
 */

/*
 *  VAI/OfoCXiSIOjhCo
 */

#include <kernel.h>
#include <t_syslog.h>
#include "target_serial.h"
#include "target_syssvc.h"

/*
 * WX^ݒl
 */
#define PORT2SIOPID(x)	((x) + 1)
#define INDEX_PORT(x)	((x) - 1)
#define GET_SIOPCB(x)	(&siopcb_table[INDEX_PORT(x)])

/*
 * UARTWX^`
 */
#define UART_SMR(x)		(x + 0x00)
#define UART_SCR(x)		(x + 0x01)
#define UART_ESCR(x)	(x + 0x04)
#define UART_SSR(x)		(x + 0x05)
#define UART_RDR(x)		(x + 0x08)
#define UART_TDR(x)		(x + 0x08)
#define UART_BGR(x)		(x + 0x0c)
#define UART_BGR0(x)	(x + 0x0c)
#define UART_BGR1(x)	(x + 0x0d)
#define UART_ISBA(x)	(x + 0x10)
#define UART_ISMK(x)	(x + 0x11)
#define UART_FCR0(x)	(x + 0x14)
#define UART_FCR1(x)	(x + 0x15)
#define UART_FBYTE1(x)	(x + 0x18)
#define UART_FBYTE2(x)	(x + 0x19)

/*
 *  VAI/O|[gubN̒`
 */
typedef struct sio_port_initialization_block {
	uint32_t base;
	INTNO    intno_rx;
	INTNO    intno_tx;
	uint16_t bps_setting;	/* {[[g̐ݒl */
} SIOPINIB;

/*
 *  pGPIO|[gubN̒`
 */
typedef struct gpio_port_initialization_block {
	uint32_t pfr;		/* PFRxWX^AhX */
	uint32_t pfr_set;	/* PFRxWX^Zbgrbg */
	uint32_t pcr;		/* PCRxWX^AhX */
	uint32_t pcr_set;	/* PCRxWX^Zbgrbg */
	uint32_t epfr;		/* EPFRxxWX^AhX */
	uint32_t epfr_clr;	/* EPFRxxWX^NArbg */
	uint32_t epfr_set;	/* EPFRxxWX^Zbgrbg */
	uint32_t ade;		/* ADEWX^AhX */
	uint32_t ade_clr;	/* ADEWX^NArbg */
} GPIOINIB;

/*
 *  VAI/O|[gǗubN̒`
 */
struct sio_port_control_block {
	const SIOPINIB  *p_siopinib;  /* VAI/O|[gubN */
	const GPIOINIB  *p_gpioinib;  /* pGPIO|[gubN */
	intptr_t        exinf;        /* g */
	bool_t          opnflg;       /* I[vς݃tO */
};

/*
 * VAI/O|[gubN
 */
const SIOPINIB siopinib_table[TNUM_SIOP] = {
	{(uint32_t)FM3_MFS4_UART_BASE, (INTNO)IRQ_VECTOR_MFS4RX, (INTNO)IRQ_VECTOR_MFS4TX, MFS4_BPS_SETTING},
	{(uint32_t)FM3_MFS0_UART_BASE, (INTNO)IRQ_VECTOR_MFS0RX, (INTNO)IRQ_VECTOR_MFS0TX, MFS0_BPS_SETTING},
	{(uint32_t)FM3_MFS3_UART_BASE, (INTNO)IRQ_VECTOR_MFS3RX, (INTNO)IRQ_VECTOR_MFS3TX, MFS3_BPS_SETTING},
};

/*
 * pGPIO|[gubN
 */
const GPIOINIB gpioinib_table[TNUM_SIOP] = {
	{(uint32_t)FM3_GPIO_PFR0, (uint32_t)((1 << 5) | (1 << 6)), (uint32_t)FM3_GPIO_PCR0, (uint32_t)(1<<5), (uint32_t)FM3_GPIO_EPFR08, (uint32_t)~0x000000f0, (uint32_t)0x000000f0, (uint32_t)FM3_GPIO_ADE, (uint32_t)0},
	{(uint32_t)FM3_GPIO_PFR2, (uint32_t)((1 << 1) | (1 << 2)), (uint32_t)FM3_GPIO_PCR2, (uint32_t)(1<<1), (uint32_t)FM3_GPIO_EPFR07, (uint32_t)~0x000000f0, (uint32_t)0x00000050, (uint32_t)FM3_GPIO_ADE, (uint32_t)(1 << 31)},
	{(uint32_t)FM3_GPIO_PFR4, (uint32_t)((1 << 8) | (1 << 9)), (uint32_t)FM3_GPIO_PCR4, (uint32_t)(1<<8), (uint32_t)FM3_GPIO_EPFR07, (uint32_t)~0x03c00000, (uint32_t)0x03c00000, (uint32_t)FM3_GPIO_ADE, (uint32_t)0},
};

/*
 *  VAI/O|[gǗubÑGA
 */
SIOPCB	siopcb_table[TNUM_SIOP];

/*
 *  VAI/O|[gIDǗubNo߂̃}N
 */
#define INDEX_SIOP(siopid)	((uint_t)((siopid) - 1))
#define get_siopcb(siopid)	(&(siopcb_table[INDEX_SIOP(siopid)]))

/*
 *  SIOhCȍ
 */
void
sio_initialize(intptr_t exinf)
{
	SIOPCB	*p_siopcb;
	uint_t	i;

	/*
	 *  VAI/O|[gǗubN̏
	 */
	for (p_siopcb = siopcb_table, i = 0; i < TNUM_SIOP; p_siopcb++, i++) {
		p_siopcb->p_siopinib = &(siopinib_table[i]);
		p_siopcb->p_gpioinib = &(gpioinib_table[i]);
		p_siopcb->opnflg = false;
	}
}


/*
 *  VAI/O|[g̃I[v
 */
SIOPCB *
sio_opn_por(ID siopid, intptr_t exinf)
{
	SIOPCB          *p_siopcb;
	const SIOPINIB  *p_siopinib;
	const GPIOINIB  *p_gpioinib;
	bool_t  opnflg;
	ER      ercd;

	p_siopcb = get_siopcb(siopid);
	p_siopinib = p_siopcb->p_siopinib;
	p_gpioinib = p_siopcb->p_gpioinib;

	/*
	 *  I[v|[g邩opnflgɓǂłD
	 */
	opnflg = p_siopcb->opnflg;

	p_siopcb->exinf = exinf;

	/*
	 *  n[hEFȀ
	 */
	sil_wrw_mem((uint32_t *)p_gpioinib->pfr, sil_rew_mem((uint32_t *)p_gpioinib->pfr) | p_gpioinib->pfr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->pcr, sil_rew_mem((uint32_t *)p_gpioinib->pcr) | p_gpioinib->pcr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->epfr, (sil_rew_mem((uint32_t *)p_gpioinib->epfr) & p_gpioinib->epfr_clr) | p_gpioinib->epfr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->ade, sil_rew_mem((uint32_t *)p_gpioinib->ade) & ~p_gpioinib->ade_clr);

	uint32_t base = p_siopinib->base;

	sil_wrb_mem((uint8_t *)UART_SCR(base), 0);
	sil_wrb_mem((uint8_t *)UART_SMR(base), SMR_MD_UART | SMR_SOE);
	sil_wrh_mem((uint16_t *)UART_BGR(base), p_siopinib->bps_setting);
	sil_wrb_mem((uint8_t *)UART_ESCR(base), ESCR_DATABITS_8);
	sil_wrb_mem((uint8_t *)UART_SCR(base), SCR_RXE | SCR_TXE);

	/*
	 *  VAI/Õ݂}XND
	 */
	if (!opnflg) {
		ercd = ena_int(p_siopinib->intno_rx);
		assert(ercd == E_OK);
		ercd = ena_int(p_siopinib->intno_tx);
		assert(ercd == E_OK);
	}

	return(p_siopcb);
}

/*
 *  VAI/O|[g̃N[Y
 */
void
sio_cls_por(SIOPCB *p_siopcb)
{
	/*
	 *  VAI/O݂}XND
	 */
	if (!(p_siopcb->opnflg)) {
		dis_int(p_siopcb->p_siopinib->intno_rx);
		dis_int(p_siopcb->p_siopinib->intno_tx);
	}
}

/*
 *  SIO̊݃T[rX[`
 */

Inline bool_t
sio_putready(SIOPCB* p_siopcb)
{
	return (sil_reb_mem((uint8_t *)UART_SSR(p_siopcb->p_siopinib->base)) & SSR_TDRE) != 0;
}

Inline bool_t
sio_getready(SIOPCB* p_siopcb)
{
	char ssr = sil_reb_mem((uint8_t *)UART_SSR(p_siopcb->p_siopinib->base));

	if ((ssr & (SSR_ORE | SSR_FRE | SSR_PE)) != 0)
	{
		sil_wrb_mem((uint8_t *)UART_SSR(p_siopcb->p_siopinib->base), ssr | SSR_REC);
		return 0;
	}
	if ((ssr & SSR_RDRF) != 0)
	{
		return 1;
	}
	return 0;
}

void
sio_tx_isr(intptr_t exinf)
{
	SIOPCB          *p_siopcb;

	p_siopcb = get_siopcb(exinf);

	if (sio_putready(p_siopcb)) {
		sio_irdy_snd(p_siopcb->exinf);
	}
}

void
sio_rx_isr(intptr_t exinf)
{
	SIOPCB          *p_siopcb;

	p_siopcb = get_siopcb(exinf);

	if (sio_getready(p_siopcb)) {
		sio_irdy_rcv(p_siopcb->exinf);
	}
}

/*
 *  VAI/O|[gւ̕M
 */
bool_t
sio_snd_chr(SIOPCB *p_siopcb, char c)
{
	if (sio_putready(p_siopcb)) {
		sil_wrh_mem((uint16_t *)UART_TDR(p_siopcb->p_siopinib->base), (uint16_t)c);

		return true;
	}

	return false;
}

/*
 *  VAI/O|[g̕M
 */
int_t
sio_rcv_chr(SIOPCB *p_siopcb)
{
	int_t c = -1;

	if (sio_getready(p_siopcb)) {
		c = sil_reh_mem((uint16_t *)UART_RDR(p_siopcb->p_siopinib->base)) & 0xFF;
	}

	return c;
}

/*
 *  VAI/O|[g̃R[obN̋
 */
void
sio_ena_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
	switch (cbrtn) {
	  case SIO_RDY_SND:
		sil_wrb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base), sil_reb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base)) | SCR_TIE);
		break;
	  case SIO_RDY_RCV:
		sil_wrb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base), sil_reb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base)) | SCR_RIE);
		break;
	}
}

/*
 *  VAI/O|[g̃R[obN̋֎~
 */
void
sio_dis_cbr(SIOPCB *p_siopcb, uint_t cbrtn)
{
	switch (cbrtn) {
	  case SIO_RDY_SND:
		sil_wrb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base), sil_reb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base)) & ~SCR_TIE);
		break;
	  case SIO_RDY_RCV:
		sil_wrb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base), sil_reb_mem((uint8_t *)UART_SCR(p_siopcb->p_siopinib->base)) & ~SCR_RIE);
		break;
	}
}

/*
 *  1ói|[Oł̏ój
 */
void sio_pol_snd_chr(char c, ID siopid)
{
	uint32_t base = siopinib_table[INDEX_PORT(siopid)].base;

	sil_wrh_mem((uint16_t *)UART_TDR(base), (uint16_t)c);
	while(0 == (sil_reb_mem((uint8_t *)UART_SSR(base)) & (1 << 1)));

	/*
	 *  o͂SɏI܂ő҂
	 */
	volatile int n = 300000000/BPS_SETTING;
	while(n--);
}

/*
 *  ^[Qbg̃VA
 */
void target_uart_init(ID siopid)
{
	const SIOPINIB  *p_siopinib = &siopinib_table[INDEX_PORT(siopid)];
	const GPIOINIB  *p_gpioinib = &gpioinib_table[INDEX_PORT(siopid)];

	sil_wrw_mem((uint32_t *)p_gpioinib->pfr, sil_rew_mem((uint32_t *)p_gpioinib->pfr) | p_gpioinib->pfr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->pcr, sil_rew_mem((uint32_t *)p_gpioinib->pcr) | p_gpioinib->pcr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->epfr, (sil_rew_mem((uint32_t *)p_gpioinib->epfr) & p_gpioinib->epfr_clr) | p_gpioinib->epfr_set);
	sil_wrw_mem((uint32_t *)p_gpioinib->ade, sil_rew_mem((uint32_t *)p_gpioinib->ade) & ~p_gpioinib->ade_clr);

	uint32_t base = siopinib_table[INDEX_PORT(siopid)].base;

	sil_wrb_mem((uint8_t *)UART_SCR(base), 0);
	sil_wrb_mem((uint8_t *)UART_SMR(base), SMR_MD_UART | SMR_SOE);
	sil_wrh_mem((uint16_t *)UART_BGR(base), p_siopinib->bps_setting);
	sil_wrb_mem((uint8_t *)UART_ESCR(base), ESCR_DATABITS_8);
	sil_wrb_mem((uint8_t *)UART_SCR(base), SCR_TXE);
}
